----------------------------------------------------------------------------
--- Fixpoint analyzer for call patterns
---
--- @author Michael Hanus
--- @version January 2008
----------------------------------------------------------------------------

import TRS
import List
import ReadFlatTRS
import Sort(mergeSort,leqString,leqList)
import Profile
import System(getArgs)
import FileGoodies(stripSuffix)
import ReadNumeric(readNat)
import TableRBT

----------------------------------------------------------------------------
-- Structure to represent abstract domains.
-- An abstract domain has the structure
--   (ADom abottom avar acons matchterms domleq)
-- and contains the following components:
-- abottom   : abstract bottom element
-- avar      : maps a free variable (number) into an abstract term
-- acons     : maps a constructor and abstract terms into an abstract term
-- matchterms: abstract matching of terms
-- domleq    : information ordering on abstract elements

data ADom a = ADom a
                   (Int -> a)
                   (String -> [a] -> a)
                   ([Term] -> [a] -> Maybe (Sub a))
                   (a -> a -> Bool)

----------------------------------------------------------------------------
-- Semantic equation (part of an interpretation): f args = result
data SemEq aterm = Eq String [aterm] aterm

-- (Abstract) Interpretation: list of semantic equations
type SemInt aterm = [SemEq aterm]

-- Substitutions: first simple implementation as string/value pairs
type Sub a = [(Int,a)]

---------------------------------------------------------------------------
-- Concrete domain: constructor terms

-- Constructor terms with bottom elements:
data CTerm = CBot | CVar Int | CCons String [CTerm]

-- is bottom term t1 a generalization (i.e., with less information) than t2?
lessCSpecific :: CTerm -> CTerm -> Bool
lessCSpecific CBot _ = True
lessCSpecific (CVar _) _ = False
lessCSpecific (CCons c1 args1) t2 = case t2 of
  CCons c2 args2 -> c1==c2 && all (uncurry lessCSpecific) (zip args1 args2)
  _              -> False

-- pairwise matching of a list of patterns against a list of terms
matchCTerms :: [Term] -> [CTerm] -> Maybe (Sub CTerm)
matchCTerms []     []     = Just []
matchCTerms []     (_:_)  = Nothing
matchCTerms (_:_)  []     = Nothing
matchCTerms (x:xs) (y:ys) = combineSubst (match x y) (matchCTerms xs ys)
 where
  combineSubst Nothing   _         = Nothing
  combineSubst (Just _ ) Nothing   = Nothing
  combineSubst (Just s1) (Just s2) = Just (s1++s2)

  -- match a linear pattern against a term with disjoint variables:
  match :: Term -> CTerm -> Maybe (Sub CTerm)
  match (Var v) t = Just [(v,t)]
  match (Func _ pf pargs) t = case t of
    CBot           -> Nothing
    CVar _         -> Nothing
    CCons tf targs -> if pf==tf then matchCTerms pargs targs else Nothing

-- The structure of the concrete domain:
concreteDom :: ADom CTerm
concreteDom = ADom CBot (error "Cannot handle free variables") CCons
                   matchCTerms lessCSpecific

-- comparison of constructor terms (only necessary for simple
-- fixpoint computation):
leqCTerm :: CTerm -> CTerm -> Bool
leqCTerm CBot _ = True
leqCTerm (CVar _) CBot = False
leqCTerm (CVar v1) (CVar v2) = v1 <= v2
leqCTerm (CVar _) (CCons _ _) = True
leqCTerm (CCons c1 args1) t2 = case t2 of
  CCons c2 args2 -> if c1==c2 then leqList leqCTerm args1 args2
                              else leqString c1 c2
  _              -> False

-- show a constructor term
showCTerm :: CTerm -> String
showCTerm CBot = "_"
showCTerm (CVar i) = 'x' : show i
showCTerm (CCons f []) = f
showCTerm (CCons f args@(_:_)) =
  f ++ "(" ++ concat (intersperse "," (map showCTerm args)) ++ ")"


----------------------------------------------------------------------------
-- Abstract domain: depth-k terms

-- depth-k terms are constructor terms with cut variables:
data DTerm = DBot | DCons String [DTerm] | CutVar

-- pairwise matching of a list of patterns against a list of terms
matchDTerms :: [Term] -> [DTerm] -> Maybe (Sub DTerm)
matchDTerms [] [] = Just []
matchDTerms [] (_:_) = Nothing
matchDTerms (_:_) [] = Nothing
matchDTerms (x:xs) (y:ys) = combineSubst (match x y) (matchDTerms xs ys)
 where
  combineSubst Nothing   _         = Nothing
  combineSubst (Just _ ) Nothing   = Nothing
  combineSubst (Just s1) (Just s2) = Just (s1++s2)

  -- match a linear pattern against a term with disjoint variables:
  match :: Term -> DTerm -> Maybe (Sub DTerm)
  match (Var v) t = Just [(v,t)]
  match (Func _ pf pargs) t = case t of
    DBot   -> Nothing
    CutVar -> Just (map (\v->(v,CutVar)) (concatMap varsOf pargs))
    DCons tf targs -> if pf==tf then matchDTerms pargs targs else Nothing

-- apply constructor on depth-k terms
consDTerm :: Int -> String -> [DTerm] -> DTerm
consDTerm maxdepth c args = cutDTerm maxdepth (DCons c args)

-- cut a depth-k term larger than a given depth:
cutDTerm :: Int -> DTerm -> DTerm
cutDTerm _ DBot = DBot
cutDTerm _ CutVar = CutVar
cutDTerm d (DCons c args) | d==0      = CutVar
                          | otherwise = DCons c (map (cutDTerm (d-1)) args)

-- is bottom term t1 a generalization (i.e., with less information) than t2?
lessDSpecific :: DTerm -> DTerm -> Bool
lessDSpecific DBot _ = True
lessDSpecific CutVar t = t==CutVar
lessDSpecific (DCons c1 args1) t2 = case t2 of
  DCons c2 args2 -> c1==c2 && all (uncurry lessDSpecific) (zip args1 args2)
  _              -> False

-- The structure of the depth-k domain (where the depth is given as an
-- argument):
depthDom :: Int -> ADom DTerm
depthDom k = ADom DBot (const CutVar) (consDTerm k) matchDTerms lessDSpecific

-- comparison of depth-k terms (only necessary for simple
-- fixpoint computation):
leqDTerm :: DTerm -> DTerm -> Bool
leqDTerm DBot _ = True
leqDTerm CutVar DBot = False
leqDTerm CutVar CutVar = True
leqDTerm CutVar (DCons _ _) = False
leqDTerm (DCons c1 args1) t2 = case t2 of
  DCons c2 args2 -> if c1==c2 then leqList leqDTerm args1 args2
                              else leqString c1 c2
  DBot           -> False
  CutVar         -> True

-- show a depth-k term
showDTerm :: DTerm -> String
showDTerm DBot = "_"
showDTerm CutVar = "*"
showDTerm (DCons f []) = f
showDTerm (DCons f args@(x:xs))
 | f == ":" && length xs == 1 = -- format list constructor:
   showDTerm x ++ ":" ++ showDTerm (head xs)
 | take 2 f == "(," = -- format tuple constructors:
   "(" ++ concat (intersperse "," (map showDTerm args)) ++ ")"
 | otherwise =
   f ++ "(" ++ concat (intersperse "," (map showDTerm args)) ++ ")"


------------------------------------------------------------------------------
-- Generic operations:

-- generic equality on interpretations (note that equations are kept ordered)
eqSemInt :: SemInt a -> SemInt a -> Bool
eqSemInt = (==)

-- generic ordered insertion of semantic equations into an interpretation
insertSemEq :: (a->a->Bool) -> SemEq a -> SemInt a -> SemInt a
insertSemEq _ x []     = [x]
insertSemEq dcmp x (y:ys) | x==y            = y : ys
                          | lessEq dcmp x y = x : y : ys
                          | otherwise       = y : insertSemEq dcmp x ys

lessEq :: (a->a->Bool) -> SemEq a -> SemEq a -> Bool
lessEq dcmp (Eq f1 args1 v1) (Eq f2 args2 v2)
  | f1==f2    = leqList dcmp (args1++[v1]) (args2++[v2])
  | otherwise = leqString f1 f2

-- generic ordered insertion of semantic equations into an interpretation,
-- but replace more general equations by specific values
-- (first argument: more general relation)
updateSemEq :: (a->a->Bool) -> (a->a->Bool) -> SemEq a -> SemInt a -> SemInt a
updateSemEq _ _ x []     = [x]
updateSemEq dcmp lessSpecific x (y:ys)
 | x == y                          = y : ys
 | lessSpecificEq lessSpecific y x = updateSemEq dcmp lessSpecific x ys
 | lessSpecificEq lessSpecific x y = y : ys
 | lessEq dcmp x y                 = x : y : ys
 | otherwise                       = y : updateSemEq dcmp lessSpecific x ys

-- is semantic equation e1 less specific (i.e., with less information)
-- than e2?
lessSpecificEq :: (a->a->Bool) -> SemEq a -> SemEq a -> Bool
lessSpecificEq lessSpecific (Eq f1 args1 v1) (Eq f2 args2 v2) =
  f1==f2 && all (uncurry lessSpecific) (zip args1 args2) && lessSpecific v1 v2

------------------------------------------------------------------------------
-- Transformation on (abstract) interpretations:
-- (transformInt matchterms applysubst evalint abottom trs mains int)
-- abottom   : abstract bottom element
-- avar      : maps a free variable (number) into an abstract term
-- acons     : maps a constructor and abstract terms into an abstract term
-- matchterms: abstract matching of terms
-- domleq    : information ordering on abstract elements
-- insertsem : operation to insert new semantic equation into an interpretation
-- trs       : term rewriting system
-- mains     : main calls (start interpretation)
-- int       : interpretation to be transformed
-- result: transformed interpretation
transformInt :: a -> (Int->a) -> (String->[a]->a)
             -> ([Term] -> [a] -> Maybe (Sub a))
             -> (a -> a -> Bool)
             -> (SemEq a -> SemInt a -> SemInt a)
             -> [Rule] -> SemInt a -> SemInt a -> SemInt a
transformInt abottom avar acons matchterms _ insertsem trs mains int =
  foldr insertsem mains
        (concatMap (\ (Eq fi argsi _) ->
                      concatMap (applyRule fi argsi) (funcRules fi trs)) int)
 where
  applyRule fi argsi (largs,rhs) =
    maybe []
          (\s -> map (Eq fi argsi) (evalInt s int rhs)
                 ++ concatMap (newSemEq s) (fSubterms rhs))
          (matchterms largs argsi)

  newSemEq s (f,args) =
    map (\iargs -> Eq f iargs abottom) (extendListMap (evalInt s int) args)

  -- abstract evaluation of a term w.r.t. a given substitution and interpretation
  evalInt :: Sub a -> SemInt a -> Term -> [a]
  evalInt sub _ (Var v) = [maybe (avar v) id (lookup v sub)]
  evalInt sub eqs (Func Cons c args) =
    map (acons c) (extendListMap (evalInt sub eqs) args)
  evalInt sub eqs (Func Def f args) =
    let evalargs = extendListMap (evalInt sub eqs) args
     in abottom :
        concatMap (\ (Eq fi argsi r) ->
              if any (\eargs->fi==f && eargs==argsi) evalargs then [r] else [])
                  eqs

-- extend list mapping on a list of elements
extendListMap :: (a -> [b]) -> [a] -> [[b]]
extendListMap _ []     = [[]]
extendListMap f (x:xs) = [ y:ys | y <- f x, ys <- extendListMap f xs]

-- compute transformation function from a given abstraction and initial calls
transformer abottom avar acons matchterms domleq insertsem rules maincalls =
  transformInt abottom avar acons matchterms domleq insertsem rules
               (foldr insertsem [] (map main2int maincalls))
 where
  -- map a main call into a abstract term
  main2int (mainf,mainargs) = Eq mainf (map cons2aterm mainargs) abottom

  -- map a constructor term into abstract term
  cons2aterm (Var v) = avar v
  cons2aterm (Func Cons c args) = acons c (map cons2aterm args)
  cons2aterm (Func Def _ _) = error "cons2aterm: Func Def occurred"


------------------------------------------------------------------------------
-- Runs a simple fixpoint computation.
runFixpoint :: ADom a
            -> (SemEq a -> SemInt a -> SemInt a)
            -> [Rule] -> [(String,[Term])] -> Bool -> (a->String)
            -> ([SemEq a] -> [SemEq a] -> Bool)
            -> IO (SemInt a)
runFixpoint (ADom abottom avar acons matchterms domleq)
            insertsem rules maincalls withprint showabs semeq = do
  let trm = transformer abottom avar acons matchterms domleq insertsem rules maincalls
  --printProgram rules maincalls
  if withprint then done else putStr "Iterating:"
  garbageCollect
  pi1 <- getProcessInfos
  fpsem <- computeFixpoint withprint 0 (showSemInt showabs) semeq trm []
  getProcessInfos >>= printTiming pi1
  return fpsem

-- Performs an iterated fixpoint computation.
computeFixpoint :: Bool -> Int -> (a->String) -> (a->a->Bool) -> (a->a) -> a
                -> IO a
computeFixpoint withprint n prt eq f v = do
  if withprint then putStrLn (show n ++ ": " ++ prt v)
               else putStr (' ':show n)
  let fv = f v
  if  eq v fv
   then do putStrLn $ "\nFixpoint reached after " ++ show n ++ " iterations"
           return v
   else computeFixpoint withprint (n+1) prt eq f fv


-- show a semantic interpretation w.r.t. a show function for semantic elements
showSemInt :: (a->String) -> SemInt a -> String
showSemInt showe eqs =
  let semIntLine = "{" ++ concat (intersperse ", " (map showEq eqs)) ++ "}"
   in show (length eqs) ++ " semantic equations:\n" ++
      if length semIntLine < 80
      then semIntLine
      else "{" ++ concat (intersperse ",\n " (map showEq eqs)) ++ "}"
 where
  showEq (Eq f args r) =
    f ++ (if null args
          then []
          else '(' : concat (intersperse "," (map showe args)) ++ ")")
      ++ " = " ++ showe r

-- get standard main call from the rules:
getMainCall :: [Rule] -> (String,[Term])
getMainCall rules = ("main", genArgs (arityOf "main" rules))
 where
  genArgs n = map Var [1..n]

-- Prints a program and a list main calls.
printProgram :: [Rule] -> [(String,[Term])] -> IO ()
printProgram rules maincalls = do
  putStrLn $ "\nRewrite rules:\n\n" ++ showTRS rules
  putStrLn $ "\nMain calls: " ++
             concat
                (intersperse ", "
                   (map (\ (f,args) -> showBasicTerm (Func Def f args))
                        maincalls))
  putStrLn ""

------------------------------------------------------------------------------
-- Fixpoint computation based on working lists.

-- The current semantic equations are represented as a mapping
-- from function names into a list of equations (arguments,results)
-- together with a list of function names on which the equation depends
type WorkSemInt a = TableRBT String [([a],a,[String])]

-- Look up the semantic equations of a function in the current semantics:
fEqsOfWorkSem :: String -> WorkSemInt a -> [([a],a,[String])]
fEqsOfWorkSem f ws = maybe [] id (lookupRBT f ws)

-- Process a working list of (abstract) calls to compute an interpretation:
-- (processWorkList adom showabs trs wl finals)
-- adom      : structure of the abstract domain
-- showabs   : show function for abstract elements (for debugging)
-- withprint : should intermediate results be printed?
-- trs       : term rewriting system
-- wl        : working list (i.e., list of abstract calls that must be
--             analyzed and inserted in final interpretation)
-- finals    : (currently) final semantic equations together with
--             a list of function names on which they depend
-- result: final interpretation

processWorkList :: ADom a -> (a -> String) -> Bool -> [Rule]
                -> [(String,[a])] -> WorkSemInt a -> IO (SemInt a)

processWorkList _ _ _ _ [] finals =
  -- transform the final semantic into the usual format:
  return (concatMap (\ (f,feqs) -> map (\ (args,res,_) -> Eq f args res) feqs)
                    (tableRBT2list finals))

processWorkList adom@(ADom abottom avar acons matchterms domleq) showabs
                withprint trs wl@((fc,eargs) : working) finals = do
  if withprint
   then putStr (" W" ++ show (length wl) ++
                "/F" ++ show (length (tableRBT2list finals)))
   else done
  let (newwss,newfeqss) = unzip (map applyRule (funcRules fc trs))
      (newws,newfeqs)   = (concat newwss, concat newfeqss)
      betterCalls       = filter isBetterCall newws
      betterEquations   = filter hasBetterResult
                                 (if null newfeqs
                                  then [(eargs,abottom,[])] -- no matching rule
                                  else newfeqs)
      bestEquations     = filter (\e-> not (any (\ei->ei/=e && leqEqDep e ei)
                                                betterEquations))
                                 betterEquations
      activatedEqs =
        if null bestEquations
        then []
        else concatMap (\ (f,feqs) -> concatMap (\ (args,_,deps) ->
                            if fc `elem` deps then [(f,args)] else []) feqs)
                       (tableRBT2list finals)
  --putStrLn ("WORKING:" ++
  --          showSemInt showabs (map (\ (f,args)->Eq f args abottom) wl))
  --putStrLn ("FINAL:" ++ showSemInt showabs
  --     (concatMap (\ (f,feqs) -> map (\ (args,res,_) -> Eq f args res) feqs)
  --                (tableRBT2list finals)))
  --putStrLn ("Function: " ++ fc)
  --putStrLn ("BETTER: " ++ show betterEquations)
  --putStrLn ("BEST  : " ++ show bestEquations)
  --putStrLn ("BETTER: " ++ show betterFuncs)
  --putStrLn ("ACTIVE: " ++ show activatedEqs)
  processWorkList adom showabs withprint trs
                 (foldr insertIfBetterCall working
                        (betterCalls ++ activatedEqs))
                 (foldr insertBetterIntoRemaining finals bestEquations)
 where
  -- insert better (dependency) equations  and delete all less specific ones:
  insertBetterIntoRemaining bettereq wsem =
    let oldfceqs = fEqsOfWorkSem fc wsem
     in updateRBT fc
          (bettereq : filter (\oldeq -> not (leqEq oldeq bettereq)) oldfceqs)
          wsem

  -- insert given abstract call if there does not already exist one
  -- which is more specific than this one; if the call is inserted,
  -- all less specific calls are removed
  insertIfBetterCall :: (String,[a]) -> [(String,[a])] -> [(String,[a])]
  insertIfBetterCall eq wlist =
    if any (leqCall eq) wlist
    then wlist
    else eq : filter (\e -> not (leqCall e eq)) wlist
   where
    leqCall (f,args) (f',args') = f==f' && leqOnList domleq args args'

  -- is a given abstract call more specific than all equations
  -- in the current final interpretation?
  isBetterCall :: (String,[a]) -> Bool
  isBetterCall (f,args) =
    not (any (\ (args',_,_) -> leqOnList domleq args args')
             (fEqsOfWorkSem f finals))

  -- is a given equation for function fc a better approximation than
  -- anything in the current final equations,
  -- i.e., if there does not already exist one final equation which has
  -- an identical left-hand side but a more specific result than this one
  hasBetterResult :: ([a],a,[String]) -> Bool
  hasBetterResult eq = not (any (leqEq eq) (fEqsOfWorkSem fc finals))

  -- compare given equation: left-hand sides and right-hand sides
  -- in leq relation on terms?
  leqEq (args,r,_) (args',r',_) = leqOnList domleq args args' && domleq r r'

  -- compare given equation and their dependencies:
  -- left- and right-hand sides in leq relation on terms and dependencies
  -- subsumed?
  leqEqDep (args,r,ds) (args',r',ds') =
     leqOnList domleq args args' && domleq r r' && all (`elem` ds') ds

  -- extend term comparison on list of terms
  leqOnList :: (a -> a -> Bool) -> [a] -> [a] -> Bool
  leqOnList _ [] [] = True
  leqOnList leq (x:xs) (y:ys) = leq x y && leqOnList leq xs ys

  applyRule (largs,rhs) =
    maybe ([],[])
          (\s -> (concatMap (newCall s) (fSubterms rhs),
                  map (\ri -> (eargs,ri,depFuncs))
                      (evalInt s finals rhs)))
          (matchterms largs eargs)
   where
    depFuncs = funcsInTerm rhs

  newCall s (f,args) = map (\iargs->(f,iargs))
                           (extendListMap (evalInt s finals) args)

  -- abstract evaluation of a term w.r.t. a substitution and interpretation
  evalInt :: Sub a -> WorkSemInt a -> Term -> [a]
  evalInt sub _ (Var v) = [maybe (avar v) id (lookup v sub)]
  evalInt sub eqs (Func Cons c args) =
    map (acons c) (extendListMap (evalInt sub eqs) args)
  evalInt sub eqs (Func Def f args) =
    let evalargs = extendListMap (evalInt sub eqs) args
        results = concatMap (\ (argsi,r,_) ->
                                  if any (\evargs->evargs==argsi) evalargs
                                  then [r]
                                  else [])
                            (fEqsOfWorkSem  f eqs)
     in if null results then [abottom] else results

-- run a fixpoint computation with working lists:
runFixpointWL :: ADom a -> [Rule] -> [(String,[Term])] -> Bool -> (a->String)
              -> IO (SemInt a)
runFixpointWL adom@(ADom _ avar acons _ _)
              rules maincalls withprint showabs = do
  --printProgram rules maincalls
  garbageCollect
  pi1 <- getProcessInfos
  finals <- processWorkList adom showabs withprint rules
                            (map main2acall maincalls)
                            (emptyTableRBT leqString)
  pi2 <- getProcessInfos
  printTiming pi1 pi2
  return finals
 where
  -- map a main call into an abstract term
  main2acall (mainf,mainargs) = (mainf, map cons2aterm mainargs)

  -- map a constructor term into abstract term
  cons2aterm (Var v) = avar v
  cons2aterm (Func Cons c args) = acons c (map cons2aterm args)
  cons2aterm (Func Def _ _) = error "cons2aterm: Func Def occurred"

-- show timing w.r.t. some process infos at the start and stop point
-- of a computation:
printTiming startPInfos stopPInfos = do
  putStrLn $ "Run time:            "
             ++ (showInfoDiff startPInfos stopPInfos RunTime) ++ " msec."
  putStrLn $ "Elapsed time:        "
             ++ (showInfoDiff startPInfos stopPInfos ElapsedTime) ++ " msec."
  putStrLn $ "Garbage collections: "
             ++ (showInfoDiff startPInfos stopPInfos GarbageCollections)
 where
  showInfoDiff infos1 infos2 item =
    maybe "n/a"
          (\i1 -> show (maybe 0 id (lookup item infos2) - i1))
          (lookup item infos1)


------------------------------------------------------------------------------
--- Main calls to the (abstract) interpreters:
------------------------------------------------------------------------------
-- main function to call the analyser as a saved state:
main = do
  args <- getArgs
  case args of
   [mainmod] -> anaCalls 2 (stripSuffix mainmod)
   ["-d",ks,mainmod] -> maybe (printErr args)
                              (\ (k,_) -> anaCalls k (stripSuffix mainmod))
                              (readNat ks)
   _ -> printErr args
 where
  anaCalls k file = do
    rules <- readRules file
    let maincall = getMainCall rules
    printProgram rules [maincall]
    fpsem <- runFixpointWL (depthDom k) rules [maincall] False showDTerm
    putStrLn (showSemInt showDTerm (mergeSort (lessEq leqDTerm) fpsem))

  printErr args = putStrLn $
    "ERROR: Illegal arguments: " ++ concat (intersperse " " args) ++ "\n" ++
    "Usage: anacalls [-d <k>] <main_module_name>\n" ++
    "       <k>: term depth (default: 2)"

------------------------------------------------------------------------------
-- a few functions for benchmarking

-- define where to look for the benchmark programs:
prog2DirFile p = "examples/"++p

-- Concrete interpretation of a program:
mc file = do
  rules <- readRules (prog2DirFile file)
  fpsem <- runFixpoint concreteDom (insertSemEq leqCTerm)
                       rules [getMainCall rules] True showCTerm eqSemInt
  putStrLn (showSemInt showCTerm fpsem)

-- Simple depth-k term interpretation of a program:
mds k file = do
  rules <- readRules (prog2DirFile file)
  fpsem <- runFixpoint (depthDom k) (insertSemEq leqDTerm)
                       rules [getMainCall rules] False showDTerm eqSemInt
  putStrLn (showSemInt showDTerm fpsem)

-- Depth-k term interpretation of a program:
md k file = do
  rules <- readRules (prog2DirFile file)
  fpsem <- runFixpoint (depthDom k) (updateSemEq leqDTerm lessDSpecific)
                       rules [getMainCall rules] False showDTerm eqSemInt
  putStrLn (showSemInt showDTerm fpsem)

-- Depth-k term interpretation of a program with working lists:
mdw k file = do
  rules <- readRules (prog2DirFile file)
  fpsem <- runFixpointWL (depthDom k) rules [getMainCall rules] False showDTerm
  putStrLn (showSemInt showDTerm (mergeSort (lessEq leqDTerm) fpsem))


-- benchmark different interpretation methods:
bench k file = do
  putStrLn (take 70 (repeat '='))
  rules <- readRules (prog2DirFile file)
  let maincall = getMainCall rules
  printProgram rules [maincall]
  putStrLn $ "\nTotal number of rules: " ++ show (length rules)
  putStrLn "\nRunning simple call analysis..."
  semsimple <- runFixpoint (depthDom k) (insertSemEq leqDTerm)
                           rules [maincall] False showDTerm eqSemInt
  putStrLn (showSemInt showDTerm semsimple)
  putStrLn "\nRunning call analysis with ordered insertion..."
  semord <- runFixpoint (depthDom k) (updateSemEq leqDTerm lessDSpecific)
                        rules [maincall] False showDTerm eqSemInt
  putStrLn (showSemInt showDTerm semord)
  putStrLn "\nRunning call analysis with working lists..."
  semwl <- runFixpointWL (depthDom k) rules [maincall] False showDTerm >>=
           return . mergeSort (lessEq leqDTerm)
  if semord==semwl
   then done
   else putStrLn (showSemInt showDTerm semwl)

-- run the benchmarks of the examples directory:
runBench = do
  let benchOp = bench --md
  benchOp 1 "addadd"
  benchOp 2 "addlast"
  benchOp 1 "bertf0"
  benchOp 1 "bertconc"
  benchOp 1 "doublecoin"
  benchOp 1 "family"
  benchOp 2 "halfdouble"
  benchOp 1 "head"
  benchOp 1 "lastapp"
  benchOp 8 "readfile"
  benchOp 2 "mapadddouble"
  benchOp 1 "risers"
  benchOp 1 "tails"
